/* set ts=8 */
/* ldapcheck.c -- Unix pwcheck daemon auth against ldap
   $Id: ldapcheck.c,v 1.13 2004/01/28 15:16:34 idra Exp $
Copyright 1998, 1999 Carnegie Mellon University
Copyright 2003 simo.sorce@xsec.it, michele@xsec.it

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#include <config.h> 

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef HAVE_PATHS_H
#include <paths.h>
#endif
#include <syslog.h>

#if !defined(_PATH_CHECKPID)
#ifdef _PATH_VARRUN
# define _PATH_CHECKPID (_PATH_VARRUN "ldapcheck.pid")
#else
# define _PATH_CHECKPID ("/var/run/ldapcheck.pid")
#endif
#endif

#define LC_SSL_OFF 0
#define LC_SSL_ON 1
#define LC_SSL_AUTO 2

#define CHECKFILE "pwcheck"

#include <ldap.h>

struct configuration {
	char *conffile;
	char *host;
	char *dn;
	char *socket;
	int port;
	int debug;
	int dom;
	int ssl;
};

struct configuration *config;

void free_configuration(struct configuration *config)
{
	if (!config) return;
	
	if (config->conffile != NULL) free(config->conffile);
	if (config->dn != NULL) free(config->dn);
	if (config->host != NULL) free(config->host);
	if (config->socket != NULL) free(config->socket);
	
	free(config);
}

void print_err(const char *fmt, ...)
{
	va_list val;
	char *errstr;
	int ret;

	va_start(val, fmt);
	ret = vasprintf(&errstr, fmt, val);
	va_end(val);
	if (ret != -1) {
		write(2, errstr, strlen(errstr)+1);
		free(errstr);
	}
}

/* replace any occurence of %u and %d in config->dn and return the resulting
 * search dn. domain may be NULL if no %d is present */
char *compose_dn(struct configuration *config, char *user, char *domain) {
	char *fdn, *p, *s, *t;
	int size, tlen, ulen, dlen;

	if (config->dom && !domain) return NULL;

	fdn = strdup(config->dn);
	if (!fdn)
		return NULL;
	
	tlen = strlen(fdn);
	ulen = strlen(user);
	if (config->dom)
		dlen = strlen(domain);

	/* search and replace */
	while ((p = strchr(fdn, '%'))) {
		int xlen;
		char *element;
		switch (p[1]) {

		case 'u':
			xlen = ulen;
			element = user;
			break;
		case 'd':
			if (config->dom) {
				xlen = dlen;
				element = domain;
			} else {
				free(fdn);
				return NULL;
			}
			break;
		default:
			continue;
		}
		size = fdn + tlen - p;
		t = (char *)realloc(fdn, tlen + xlen + 1);
		if (!t) {
			free(fdn);
			return NULL;
		}
		s = &t[p - fdn];
		fdn = t;
		memmove(&s[xlen], &s[2], size);
		if (xlen) memcpy(s, element, xlen);
		tlen = strlen(fdn);
	}

	return fdn;
}
		
/* Allocates a configuration struct and fills it in with values parsed in the
 * conffile FIXME: parser needs to be more robust*/
int get_configuration(struct configuration *config)
{
	FILE *fd;
	int debug = config->debug;

	fd = fopen(config->conffile, "r");
	if (fd == NULL) { /* Unable to open the conffile, fail */
		print_err("Unable to open %s\n", config->conffile);
		return -1;
	}

	while (!feof(fd)) {
		char confline[4096] = "";
		char *option, *value;
		int c, clen, vlen;

		if (fgets(confline, 4096, fd) == NULL) continue;

		clen = strlen(confline);
		if (confline[clen - 1] == '\n') {
			confline[clen - 1] = '\0';
			clen--;
		}
		if (clen == 0) continue;
		
		/* strip leading white chars and tabs */
		for (c = 0; c < clen; c++) {
			if (confline[c] != ' ' && confline[c] != '\t')
				break;
		}
		if (c == clen) continue;
		if (confline[c] == '#') continue;

		option = &confline[c];

		value = strchr(option, ' ');
		if (value == NULL) {
			print_err("Unknown option: %s\n", confline);
			continue;
		}
		*value = '\0';
		value++;

		/* strip trailing white chars, tabs and newlines from */
		for (c = strlen(value) - 1; c >= 0; c--) {
			if (value[c] != ' ' &&	value[c] != '\t')
				break;
		}
		if (c == 0) {
			print_err("Malformed option: %s\n", confline);
			continue;
		}
		value[c + 1] = '\0';
		vlen = strlen(value);

		if (strcmp(option, "dn") == 0) {
			config->dn = (char *)malloc(vlen + 1);	
			strcpy(config->dn, value);
		}
		else if (strcmp(option, "host") == 0) {
			config->host = (char *)malloc(vlen + 1);	
			strcpy(config->host, value);
		}
		else if (strcmp(option, "debug") == 0) {
			if (strcmp(value, "on") == 0) config->debug = 1;
			else config->debug = 0;
		}
		else if (strcmp(option, "ssl") == 0) {
			config->ssl = LC_SSL_AUTO;
			if (strcmp(value, "off") == 0) config->ssl = LC_SSL_OFF;
			else if (strcmp(value, "on")) config->ssl = LC_SSL_ON;
		}
		else if (strcmp(option, "socket") == 0 ||
				strcmp(option, "socketfile") == 0) {
			config->socket = (char *)malloc(vlen + 1);
			strcpy(config->socket, value);
		}
		else if (strcmp(option, "port") == 0) { 
			config->port = atoi(value);
		}
		else {
			print_err("Unknown option: %s\n", confline);
			continue;
		}
		if (debug) print_err("%s -> %s\n", option, value);
	}
	fclose(fd);

	/* Set some sensible defaults, if no values where specified */
	if (config->port == 0) config->port = 389;
	if (config->host == NULL) config->host = strdup("localhost");

	if (!(config->dn)) {
		print_err("No dn in config file! Bailing out...\n");
		return -1;
	}

	if (config->socket == NULL) {
		config->socket = (char *)malloc(sizeof(CHECKDIR)+1+sizeof(CHECKFILE)+1);
		strcpy(config->socket, CHECKDIR);
		strcat(config->socket, "/");
		strcat(config->socket, CHECKFILE);
	}

	/* check if we expect a domain part in username */
	if (strstr(config->dn, "%d")) {
		config->dom = 1;
	} else {
		config->dom = 0;
	}

	return 0;
}

void usage(void)
{
	print_err( 
		"Usage: ldapcheck [ Options ] \n"
		"         -d         - Enable debug option\n"
		"         -c <file>  - Configuration file (unless specified it is /etc/ldapcheck.conf)\n"
		"         -v         - Print out version\n"
		"         -h         - Print this help\n");
}

char *ldapcheck(char *username, char *password)
{
	char *result = "Authentication Failed";
	char *user = NULL;
	char *dn = NULL;
	char *p;
	LDAP *handle = NULL;
	int ret, dom = 0;
	
	if (username == NULL || *username == '\0' || strchr(username, '%') || 
		password == NULL || *password == '\0') {
		/* We do not accept null user or null password auths */
		goto done;
	}

	if (config->debug) syslog(LOG_INFO, "check username: %s", username);

	handle = ldap_init(config->host, config->port);
	if (!handle) {
		syslog(LOG_ERR, "ldap_init failed!");
		goto done;
	}

	user = strdup(username);
	if (!user) {
		syslog(LOG_ERR, "Out of memory!");
		goto done;
	}

	p = strchr(user, (int)'@');

	if (p && *(p + 1) != '\0') {
		dom = 1;
		*p = '\0';
		p++;
	}

	if (config->dom && !dom) {
		if (config->debug) syslog(LOG_ERR, "Domain not provided");
		goto done;
	}

	if (dom) {
		dn = compose_dn(config, user, p);
	} else {
		dn = compose_dn(config, user, NULL);
	}
	if (dn == NULL) {
		syslog(LOG_ERR, "Internal Error! Unable to properly build dn");
		goto done;
	}

	if (config->debug) syslog(LOG_INFO, "Checking dn: %s", dn);

	ret = ldap_simple_bind_s(handle, dn, password);
	if (ret != LDAP_SUCCESS) {
		if (config->debug) syslog(LOG_INFO, "Auth failed: %d", ret);
		goto done;
	}

	if (config->debug) syslog(LOG_INFO, "Auth succeded");
	result = "OK";

done:
	if (handle) ldap_unbind_s(handle);
	if (user) free(user);
	if (dn) free(dn);
	return result;
}
  
/*
 * Keep calling the write() system call with 'fd', 'buf', and 'nbyte'
 * until all the data is written out or an error occurs.
 */

int retry_write(int fd, const char *buf, unsigned int nbyte)
{
	int n;
	int written = 0;

	if (nbyte == 0)
		return 0;

	for (;;) {
		n = write(fd, buf, nbyte);
		if (n == -1) {
			if (errno == EINTR)
				continue;
			return -1;
   	}

		written += n;

		if ((unsigned int) n >= nbyte)
			return written;

		buf += n;
		nbyte -= n;
	}
}

void newclient(int c)
{
	char request[1024];
	int n;
	unsigned int start;
	char *reply;
    
	start = 0;
	while (start < sizeof(request) - 1) {
		n = read(c, request+start, sizeof(request) - 1 - start);
		if (n < 1) {
			reply = "Error reading request";
			goto sendreply;
		}

		start += n;
		if (request[start-1] == '\0' && strlen(request) < start) {
			break;
		}
	}

	if (start >= sizeof(request) - 1) {
		reply = "Request too big";
	} else {
		reply = ldapcheck(request, request + strlen(request) + 1);
	}

sendreply:

	retry_write(c, reply, strlen(reply));
	close(c);
}


int main(int argc, char *argv[])
{
	int s;
	int c;
	int count;
	int rc;
	struct sockaddr_un srvaddr;
	struct sockaddr_un clientaddr;
	int r;
	int len;
	mode_t oldumask;
	char *pid_file = _PATH_CHECKPID;
	FILE *fp = NULL;
	pid_t pid;
	int tmp_debug = 0;

	config = (struct configuration *)malloc(sizeof(struct configuration));
	memset(config, 0, sizeof(struct configuration));	

	while (argc > 1) {
		char *opt = argv[1];
		if (strcmp(opt, "-h") == 0) {
			usage();
			free_configuration(config);
			exit(-1);
		}
		else if (strcmp(opt, "-v") == 0) {
			printf("ldapcheck version %s\n", VERSION);
			exit(0);
		}
		else if (strcmp(opt, "-d") == 0) {
			tmp_debug = 1;
			config->debug = 1;
		}
		else if (strcmp(opt, "-c") == 0) {
			argc--; argv++;
			if (argc == 0) {
				usage();
				free_configuration(config);
				exit(-1);
			}
			opt = argv[1];
			config->conffile = (char *)malloc(strlen(opt));
			strcpy(config->conffile, opt);
		}
		else {
			print_err("Unknown option [%s]\n!", opt);
			usage();
			free_configuration(config);
			exit(-1);
		}
		argc--; argv++;
	}

	if (config->conffile == NULL) {
		config->conffile = (char *)malloc(strlen(CONFIGFILE) + 1);
		strcpy(config->conffile, CONFIGFILE);
	}
	
	if (get_configuration(config) != 0) {
		free_configuration(config);
		exit(-1);
	}

	/* Set the debug level as to what the user decided on the command line, no
	 * matter what */
	config->debug = tmp_debug;

	openlog("ldapcheck", LOG_NDELAY, LOG_AUTH);
	/* Daemonize. */
	count = 5;
	while (count--) {
		pid = fork();
		if (pid > 0) _exit(0);               /* parent dies */
            
		if ((pid == -1) && (errno == EAGAIN)) {
			syslog(LOG_WARNING, "master fork failed (sleeping): %m");
			sleep(5);
			continue;
		}
	}
		
	if (pid == -1) {
		rc = errno;
		syslog(LOG_ERR, "FATAL: master fork failed: %m");
		fprintf(stderr, "ldapcheck: ");
		errno = rc;
		perror("fork");
		exit(1);
	}

	/*
	 * We're now running in the child. Lose our controlling terminal
	 * and obtain a new process group.
	 */
	if (setsid() == -1) {
		rc = errno;
		syslog(LOG_ERR, "FATAL: setsid: %m");
		fprintf(stderr, "ldapcheck: ");
		errno = rc;
		perror("setsid");
		exit(1);
	}
        
	s = open("/dev/null", O_RDWR, 0);
	if (s == -1) {
		rc = errno;
		syslog(LOG_ERR, "FATAL: /dev/null: %m");
		fprintf(stderr, "ldapcheck: ");
		errno = rc;
		perror("/dev/null");
		exit(1);
	}

	dup2(s, fileno(stdin));
	dup2(s, fileno(stdout));
	dup2(s, fileno(stderr));
	if (s > 2) {
		close(s);
	}

	/*
	 *   Record process ID - shamelessly stolen from inetd (I.V.)
	 */
	pid = getpid();
	if (pid_file) {
		fp = fopen(pid_file, "w");
	}
	if (fp) {
		fprintf(fp, "%ld\n", (long)pid);
		fclose(fp);
	} else if (pid_file) {
		syslog(LOG_WARNING, "%s: %m", pid_file);
		exit(-1);
	}

	s = socket(AF_UNIX, SOCK_STREAM, 0);
	if (s == -1) {
		perror("socket");
		exit(-1);
	}

	(void) unlink(config->socket);

	memset((char *)&srvaddr, 0, sizeof(srvaddr));
	srvaddr.sun_family = AF_UNIX;
	strncpy(srvaddr.sun_path, config->socket, sizeof(srvaddr.sun_path));
	/* Most systems make sockets 0777 no matter what you ask for.
	   Known exceptions are Linux and DUX. */
	oldumask = umask((mode_t) 0); /* for Linux, which observes the umask when setting up the socket */
	r = bind(s, (struct sockaddr *)&srvaddr, sizeof(srvaddr));
	if (r == -1) {
		syslog(LOG_ERR, "%.*s: %m", sizeof(srvaddr.sun_path), srvaddr.sun_path);
		exit(-1);
	}
	umask(oldumask); /* for Linux */
       	/* for DUX, where this isn't the default.
	 * (harmlessly fails on some systems) */	
	chmod(config->socket, (mode_t) 0777);

	r = listen(s, 5);
	if (r == -1) {
		syslog(LOG_ERR, "listen: %m");
		exit(-1);
	}

	for (;;) {
		len = sizeof(clientaddr);
		c = accept(s, (struct sockaddr *)&clientaddr, &len);
		if (c == -1 && errno != EINTR) {
			syslog(LOG_WARNING, "accept: %m");
			continue;
		}
		if (config->debug) syslog(LOG_INFO, "accept: %m");
		newclient(c);
	}
}
